// Rebecca Fiebrink
// www.cs.princeton.edu/~fiebrink/
// Example of using Kinect for Wekinator feature extractor, with skeleton tracking
// Extracts 69 features:  3D position for each of 23 joints (not all joints may be updated by Kinect
// Adapted from Stickmanetic: https://github.com/Sensebloom/OSCeleton-examples
// 

// Requires oscP5, and netP5 to be installed and working. (And a Kinect, of course!)

import oscP5.*;
import netP5.*;

import pbox2d.*;
import org.jbox2d.collision.shapes.*;
import org.jbox2d.common.*;
import org.jbox2d.dynamics.*;

// A reference to our box2d world
PBox2D box2d;
OscP5 oscP5;

float noiseIter = 0.0;
int ballSize = 10;

Hashtable<Integer, Skeleton> skels = new Hashtable<Integer, Skeleton>();

NetAddress dest;
float[] features = new float[23*3];

void setup() {
  size(640, 480);    // use OPENGL rendering for bilinear filtering on texture
  smooth();
    
  // Initialize box2d physics and create the world
  box2d = new PBox2D(this);
  box2d.createWorld();
  // We are setting a custom gravity
  box2d.setGravity(0, -40);
  
  //Listen on port 7110 for joint info from osceleton
  oscP5 = new OscP5(this, 7110);
  dest = new NetAddress("127.0.0.1",6448); //Send to Wekinator at 6448 (don't change), running on local host (could change host)
  sendNames(); //send feature names so they're displayed nicely in Wekinator GUI
}

//Send feature names to Wekinator so they're displayed nicely in GUI
//Only has to be called once (after Wekinator is launched), but it doesn't hurt if it's called more than once
void sendNames() {
  OscMessage msg = new OscMessage("/oscCustomFeaturesNames");
  String[] parts = {"head", "neck", "rcollar", "rshoulder", "relbow", "rwrist", "rhand", "rfinger", "lcollar", "lshoulder", "lelbow", "lwrist", "lhand", "lfinger", "torso", "rhip", "rknee", "rankle", "rfoot", "lhip", "lknee", "lankle", "rfoot"};
  for (int p = 0; p < parts.length; p++) {
    msg.add(parts[p] + "_x");
    msg.add(parts[p] + "_y");
    msg.add(parts[p] + "_z");
  }
  oscP5.send(msg, dest); 
}

// Processes incoming OSC messages from osceleton
// Here you can easily see the format of the OSC messages sent. For each user, the joints are named with 
// the joint named followed by user ID (head0, neck0 .... r_foot0; head1, neck1.....)
void oscEvent(OscMessage msg) {
  if (msg.checkAddrPattern("/joint") && msg.checkTypetag("sifff")) {
    // We have received joint coordinates, let's find out which skeleton/joint and save the values ;)
    Integer id = msg.get(1).intValue();
    Skeleton s = skels.get(id);
    if (s == null) {
      s = new Skeleton(id);
      skels.put(id, s);
    }
    if (msg.get(0).stringValue().equals("head")) {
      s.headCoords[0] = msg.get(2).floatValue();
      s.headCoords[1] = msg.get(3).floatValue();
      s.headCoords[2] = msg.get(4).floatValue();
      features[0] = s.headCoords[0];
      features[1] = s.headCoords[1];
      features[2] = s.headCoords[2];
    }
    else if (msg.get(0).stringValue().equals("neck")) {
      s.neckCoords[0] = msg.get(2).floatValue();
      s.neckCoords[1] = msg.get(3).floatValue();
      s.neckCoords[2] = msg.get(4).floatValue();
      features[3] = s.neckCoords[0];
      features[4] = s.neckCoords[1];
      features[5] = s.neckCoords[2];
    }
    else if (msg.get(0).stringValue().equals("r_collar")) {
      s.rCollarCoords[0] = msg.get(2).floatValue();
      s.rCollarCoords[1] = msg.get(3).floatValue();
      s.rCollarCoords[2] = msg.get(4).floatValue();
      features[6] = s.rCollarCoords[0];
      features[7] = s.rCollarCoords[1];
      features[8] = s.rCollarCoords[2];
    }
    else if (msg.get(0).stringValue().equals("r_shoulder")) {
      s.rShoulderCoords[0] = msg.get(2).floatValue();
      s.rShoulderCoords[1] = msg.get(3).floatValue();
      s.rShoulderCoords[2] = msg.get(4).floatValue();
      features[9] = s.rShoulderCoords[0];
      features[10] = s.rShoulderCoords[1];
      features[11] = s.rShoulderCoords[2];
    }
    else if (msg.get(0).stringValue().equals("r_elbow")) {
      s.rElbowCoords[0] = msg.get(2).floatValue();
      s.rElbowCoords[1] = msg.get(3).floatValue();
      s.rElbowCoords[2] = msg.get(4).floatValue();
       features[12] = s.rElbowCoords[0];
      features[13] = s.rElbowCoords[1];
      features[14] = s.rElbowCoords[2];
    }
    else if (msg.get(0).stringValue().equals("r_wrist")) {
      s.rWristCoords[0] = msg.get(2).floatValue();
      s.rWristCoords[1] = msg.get(3).floatValue();
      s.rWristCoords[2] = msg.get(4).floatValue();
      features[15] = s.rWristCoords[0];
      features[16] = s.rWristCoords[1];
      features[17] = s.rWristCoords[2];
    }
    else if (msg.get(0).stringValue().equals("r_hand")) {
      s.rHandCoords[0] = msg.get(2).floatValue();
      s.rHandCoords[1] = msg.get(3).floatValue();
      s.rHandCoords[2] = msg.get(4).floatValue();
      features[18] = s.rHandCoords[0];
      features[19] = s.rHandCoords[1];
      features[20] = s.rHandCoords[2];
    }
    else if (msg.get(0).stringValue().equals("r_finger")) {
      s.rFingerCoords[0] = msg.get(2).floatValue();
      s.rFingerCoords[1] = msg.get(3).floatValue();
      s.rFingerCoords[2] = msg.get(4).floatValue();
       features[21] = s.rFingerCoords[0];
      features[22] = s.rFingerCoords[1];
      features[23] = s.rFingerCoords[2];
    }
    else if (msg.get(0).stringValue().equals("l_collar")) {
      s.lCollarCoords[0] = msg.get(2).floatValue();
      s.lCollarCoords[1] = msg.get(3).floatValue();
      s.lCollarCoords[2] = msg.get(4).floatValue();
      features[24] = s.lCollarCoords[0];
      features[25] = s.lCollarCoords[1];
      features[26] = s.lCollarCoords[2];
    }  
    else if (msg.get(0).stringValue().equals("l_shoulder")) {
      s.lShoulderCoords[0] = msg.get(2).floatValue();
      s.lShoulderCoords[1] = msg.get(3).floatValue();
      s.lShoulderCoords[2] = msg.get(4).floatValue();
      features[27] = s.lShoulderCoords[0];
      features[28] = s.lShoulderCoords[1];
      features[29] = s.lShoulderCoords[2];
    }
    else if (msg.get(0).stringValue().equals("l_elbow")) {
      s.lElbowCoords[0] = msg.get(2).floatValue();
      s.lElbowCoords[1] = msg.get(3).floatValue();
      s.lElbowCoords[2] = msg.get(4).floatValue();
       features[30] = s.lElbowCoords[0];
      features[31] = s.lElbowCoords[1];
      features[32] = s.lElbowCoords[2];
    }
    else if (msg.get(0).stringValue().equals("l_wrist")) {
      s.lWristCoords[0] = msg.get(2).floatValue();
      s.lWristCoords[1] = msg.get(3).floatValue();
      s.lWristCoords[2] = msg.get(4).floatValue();
      features[33] = s.lWristCoords[0];
      features[34] = s.lWristCoords[1];
      features[35] = s.lWristCoords[2];
    }
    else if (msg.get(0).stringValue().equals("l_hand")) {
      s.lHandCoords[0] = msg.get(2).floatValue();
      s.lHandCoords[1] = msg.get(3).floatValue();
      s.lHandCoords[2] = msg.get(4).floatValue();
      features[36] = s.lHandCoords[0];
      features[37] = s.lHandCoords[1];
      features[38] = s.lHandCoords[2];
    }
    else if (msg.get(0).stringValue().equals("l_finger")) {
      s.lFingerCoords[0] = msg.get(2).floatValue();
      s.lFingerCoords[1] = msg.get(3).floatValue();
      s.lFingerCoords[2] = msg.get(4).floatValue();
      features[39] = s.lFingerCoords[0];
      features[40] = s.lFingerCoords[1];
      features[41] = s.lFingerCoords[2];
    }
    else if (msg.get(0).stringValue().equals("torso")) {
      s.torsoCoords[0] = msg.get(2).floatValue();
      s.torsoCoords[1] = msg.get(3).floatValue();
      s.torsoCoords[2] = msg.get(4).floatValue();
       features[42] = s.torsoCoords[0];
      features[43] = s.torsoCoords[1];
      features[44] = s.torsoCoords[2];
    }
    else if (msg.get(0).stringValue().equals("r_hip")) {
      s.rHipCoords[0] = msg.get(2).floatValue();
      s.rHipCoords[1] = msg.get(3).floatValue();
      s.rHipCoords[2] = msg.get(4).floatValue();
       features[45] = s.rHipCoords[0];
      features[46] = s.rHipCoords[1];
      features[47] = s.rHipCoords[2];
    } 
    else if (msg.get(0).stringValue().equals("r_knee")) {
      s.rKneeCoords[0] = msg.get(2).floatValue();
      s.rKneeCoords[1] = msg.get(3).floatValue();
      s.rKneeCoords[2] = msg.get(4).floatValue();
       features[48] = s.rKneeCoords[0];
      features[49] = s.rKneeCoords[1];
      features[50] = s.rKneeCoords[2];
    } 
    else if (msg.get(0).stringValue().equals("r_ankle")) {
      s.rAnkleCoords[0] = msg.get(2).floatValue();
      s.rAnkleCoords[1] = msg.get(3).floatValue();
      s.rAnkleCoords[2] = msg.get(4).floatValue(); 
      features[51] = s.rAnkleCoords[0];
      features[52] = s.rAnkleCoords[1];
      features[53] = s.rAnkleCoords[2];
    } 
    else if (msg.get(0).stringValue().equals("r_foot")) {
      s.rFootCoords[0] = msg.get(2).floatValue();
      s.rFootCoords[1] = msg.get(3).floatValue();
      s.rFootCoords[2] = msg.get(4).floatValue();
       features[54] = s.rFootCoords[0];
      features[55] = s.rFootCoords[1];
      features[56] = s.rFootCoords[2];
    } 
    else if (msg.get(0).stringValue().equals("l_hip")) {
      s.lHipCoords[0] = msg.get(2).floatValue();
      s.lHipCoords[1] = msg.get(3).floatValue();
      s.lHipCoords[2] = msg.get(4).floatValue();
      features[57] = s.lHipCoords[0];
      features[58] = s.lHipCoords[1];
      features[59] = s.lHipCoords[2];
    } 
    else if (msg.get(0).stringValue().equals("l_knee")) {
      s.lKneeCoords[0] = msg.get(2).floatValue();
      s.lKneeCoords[1] = msg.get(3).floatValue();
      s.lKneeCoords[2] = msg.get(4).floatValue();
      features[60] = s.lKneeCoords[0];
      features[61] = s.lKneeCoords[1];
      features[62] = s.lKneeCoords[2];
    } 
    else if (msg.get(0).stringValue().equals("l_ankle")) {
      s.lAnkleCoords[0] = msg.get(2).floatValue();
      s.lAnkleCoords[1] = msg.get(3).floatValue();
      s.lAnkleCoords[2] = msg.get(4).floatValue();
      features[63] = s.lAnkleCoords[0];
      features[64] = s.lAnkleCoords[1];
      features[65] = s.lAnkleCoords[2];
    } 
    else if (msg.get(0).stringValue().equals("l_foot")) {
      s.lFootCoords[0] = msg.get(2).floatValue();
      s.lFootCoords[1] = msg.get(3).floatValue();
      s.lFootCoords[2] = msg.get(4).floatValue();
      features[66] = s.lFootCoords[0];
      features[67] = s.lFootCoords[1];
      features[68] = s.lFootCoords[2];
    } 
  }
  else if (msg.checkAddrPattern("/new_user") && msg.checkTypetag("i")) {
    // A new user is in front of the kinect... Tell him to do the calibration pose!
    println("New user with ID = " + msg.get(0).intValue());
  }
  else if(msg.checkAddrPattern("/new_skel") && msg.checkTypetag("i")) {
    //New skeleton calibrated! Lets create it!
    Integer id = msg.get(0).intValue();
    Skeleton s = new Skeleton(id);
    skels.put(id, s);
  }
  else if(msg.checkAddrPattern("/lost_user") && msg.checkTypetag("i")) {
    //Lost user/skeleton
    Integer id = msg.get(0).intValue();
    println("Lost user " + id);
    skels.remove(id);
  }
}


void drawBone(float joint1[], float joint2[]) {
  if ((joint1[0] == -1 && joint1[1] == -1) || (joint2[0] == -1 && joint2[1] == -1))
    return;
  
  float dx = (joint2[0] - joint1[0]) * width;
  float dy = (joint2[1] - joint1[1]) * height;
  float steps = 2 * sqrt(pow(dx,2) + pow(dy,2)) / ballSize;
  float step_x = dx / steps / width;
  float step_y = dy / steps / height;
  
  for (int i=0; i<=steps; i++) {
    ellipse((joint1[0] + (i*step_x))*width, 
            (joint1[1] + (i*step_y))*height, 
            ballSize, ballSize);
  }
}


void draw() {
  background(0);
  noStroke();

  
  for (Skeleton s: skels.values()) {
    s.addCollisionLine();
  
    //Head
    ellipse(s.headCoords[0]*width, 
            s.headCoords[1]*height + 10, 
            ballSize*5, ballSize*6);
  
    //Head to neck 
    drawBone(s.headCoords, s.neckCoords);
    //Center upper body
    //drawBone(lShoulderCoords, rShoulderCoords);
    drawBone(s.headCoords, s.rShoulderCoords);
    drawBone(s.headCoords, s.lShoulderCoords);
    drawBone(s.neckCoords, s.torsoCoords);
    //Right upper body
    drawBone(s.rShoulderCoords, s.rElbowCoords);
    drawBone(s.rElbowCoords, s.rHandCoords);
    //Left upper body
    drawBone(s.lShoulderCoords, s.lElbowCoords);
    drawBone(s.lElbowCoords, s.lHandCoords);
    //Torso
    //drawBone(rShoulderCoords, rHipCoords);
    //drawBone(lShoulderCoords, lHipCoords);
    drawBone(s.rHipCoords, s.torsoCoords);
    drawBone(s.lHipCoords, s.torsoCoords);
    //drawBone(lHipCoords, rHipCoords);
    //Right leg
    drawBone(s.rHipCoords, s.rKneeCoords);
    drawBone(s.rKneeCoords, s.rFootCoords);
  //  drawBone(rFootCoords, lHipCoords);
    //Left leg
    drawBone(s.lHipCoords, s.lKneeCoords);
    drawBone(s.lKneeCoords, s.lFootCoords);
  //  drawBone(lFootCoords, rHipCoords);
    
    for (float j[]: s.allCoords) {
      ellipse(j[0]*width, j[1]*height, ballSize*2, ballSize*2);
    }  
    s.body.createShape(s.edges);
  }
  
  noiseIter += 0.01;
  
  // We must always step through time!
  box2d.step();


  for (Skeleton s: skels.values()) {
    box2d.destroyBody(s.body);
    s.body = box2d.world.createBody(s.bd);
    s.edges = new EdgeChainDef();
    s.edges.setIsLoop(false);   // We could make the edge a full loop
    s.edges.friction = 0.1;    // How much friction
    s.edges.restitution = 1.3; // How bouncy
  }
  
  // Concatenate all joint features and send to Wekinator
  // if(frameCount % 5 == 0) { //Could optionally send less frequently to be more efficient
    sendOsc();
  // }
}

//Send all features to Wekinator as a single vector of floats
void sendOsc() {
  OscMessage msg = new OscMessage("/oscCustomFeatures");
  for (int i= 0; i < features.length; i++){
      msg.add(features[i]);
  }
  oscP5.send(msg, dest);
}
